'' =================================================================================================
''
''   File....... hc8_stranger-things_demo__17-07-26.spin 
''   Purpose.... Demo program for EFX-TEK HC-8+ controller
''               -- translate string intput to outputs on WS2812 string
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (C) 2011-2017 Jon McPhalen
''   E-mail..... 
''   Started.... 
''   Updated.... 26 JUL 2017 - modified to allow other than Rev F boards to run
''
'' =================================================================================================


con

  ' Important: Set Rev letter to match your baord. Rev E and lower boards do not have
  '            XPort pins; XPort functions are disabled for Rev E and lower.

  PCB_REV = "F"


con { timing }

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000                                          ' use 5MHz crystal

  CLK_FREQ = _xinfreq * (_clkmode >> 6)                         ' system freq as a constant
  MS_001   = CLK_FREQ / 1_000                                   ' ticks in 1ms
  US_001   = CLK_FREQ / 1_000_000                               ' ticks in 1us
                                                                 
  USB_BAUD = 115_200                                            ' standard PST baud
  BT_BAUD  =  57_600                                            ' bluetooth 
                                                                 
                                                                 
con { io pins }                                                  
                                                                 
  RX1     = 31  { I }                                           ' serial / programming
  TX1     = 30  { O }                                            
                                                                 
  SDA     = 29  { I/O }                                         ' i2c / eeprom
  SCL     = 28  { I/O }                                          
                                                                 
  XP0     = 27  { I/O }                                         ' extra port 0 - Rev F
  XP1     = 24  { I/O }                                         ' extra port 1 - Rev F
  XP2     = 19  { I/O }                                         ' extra port 2 - Rev F  
  XP3     = 16  { I/O }                                         ' extra port 3 - Rev F

  OUT_EN  = 27  { O }                                           ' outputs /enable - Rev E & lower     
  
  SIO     = 26  { I/O }                                         ' TTL serial io - Rev E & lower
                                                          
  SIO_RX  = 26  { I/O }                                         ' TTL serial io - Rev F   
  SIO_TX  = 25  { I/O }                                         ' TTL serial io - Rev F  (not enabled)  
                                                                 
  RX2     = 23  { I }                                           ' RS-485 port
  RX2_EN  = 22  { O }                                                
  TX2_EN  = 21  { O }                                                
  TX2     = 20  { O }                                                
  
  G_LED   = 18  { O }                                           ' bi-color LED (Rev F assignments)
  R_LED   = 17  { O }                                              
                                                                 
  OUT7    = 15  { O }                                           ' outputs
  OUT6    = 14  { O }                                                
  OUT5    = 13  { O }                                                
  OUT4    = 12  { O }                                                
  OUT3    = 11  { O }                                                
  OUT2    = 10  { O }                                                
  OUT1    =  9  { O }                                                
  OUT0    =  8  { O }                                                
                                                                 
  OPT_BR  =  7  { I/O }                                         ' options switches
  OPT_A1  =  6  { I/O }                                                 
  OPT_A0  =  5  { I/O }                                                 
  OPT_SM  =  4  { I/O }                                                 
                                                                 
  DMX_A8  =  3  { I }                                           ' high-bit of DMX address
                                                                 
  DO_165  =  2  { I }                                           ' control lines for 74x165
  CLK_165 =  1  { O }                                                 
  LD_165  =  0  { O }                                                 


con { app connections }

  RX3     = OPT_A0  { I }                                       ' for Bluetooth module
  TX3     = OPT_SM  { O }


con { serial connection for remote commands }

  #0, M_USB, M_RADIO                                            ' serial connection modes

  CMD_MODE = M_USB                                              ' command mode
               
                                                                 
con                                                              
                                                                 
  CHANNELS = 7                                                  ' protect OUT7 (for WS2812s)                                                   
  LAST_CH  = CHANNELS - 1                                        

  STRIP_LEN = 38                                                ' # of bulbs in strip
                                                                
                                                                 
  #(true), ON, OFF                                              ' digital control
                                                                 
  #0, NO, YES                                                    
                                                                 
  #0, BLK, RED, GRN, YEL                                        ' for r/g led
                                                                 
  #0, NOCON, RX, TX, RXTX                                       ' for rs-485
                                                                 
  #0, LSBFIRST, MSBFIRST                                         
                                                                 
                                                                 
obj                              
                                 
' main                                                          ' * master Spin cog                   
  time       : "jm_time_80"                                     '   timing and delays (80MHz system) 
  prng       : "jm_prng"                                        '   pseudo-random number generator
  text       : "jm_textbuf"                                     '   text buffer manipulation     
  io         : "jm_io"                                          '   essential io  
  outs       : "jm_pwm8"                                        ' * pwm driver for outputs
  term       : "jm_fullduplexserial"                            ' * serial IO for terminal
  strip      : "jm_rgbx_pixel"                                  ' * ws2812 control
' background                                                    ' * background RGY LED control, inputs scanning 

' * uses cog when loaded
                                 
                                 
var

  long  pixels1[STRIP_LEN]                                      ' primary pixels buffer
  long  pixels2[STRIP_LEN]                                      ' secondary pixels buffer

  byte  tbuf[40]                                                ' incoming text buffer


con

  #0, C_RED, C_YEL, C_GRN, C_CYN, C_BLU, C_VIO                  ' base bulb colors
    

dat

  Colors                byte    6

  ColorTable            long                                    ' correct alignment
  ST_Red                long    $80_00_00_00                    ' red
  ST_Yellow             long    $50_38_00_00                    ' yellow
  ST_Green              long    $00_80_00_00                    ' green
  ST_Cyan               long    $00_42_42_00                    ' cyan
  ST_Blue               long    $00_00_80_00                    ' blue
  ST_Violet             long    $40_00_40_00                    ' violet

  OtherColors           long           
  ST_Orange             long    $50_10_00_00
  ST_Purple             long    $30_00_60_00
  ST_Pink               long    $60_20_20_00
  ST_White              long    $2A_2A_2A_00
            
                                                                 
  ' positions of letters in string display
  ' -- table contains WS281x channel # for letter
  
                        '        a   b   c   d   e   f   g   h                        
  Alphas  { A..Z }      byte     0,  1,  2,  3,  4,  5,  6,  7
                        '        i   j   k   l   m   n   o   p 
                        byte    21, 19, 18, 17, 16, 15, 14, 12
                        '        q   r   s   t   u   v   w   x 
                        byte    10, 25, 26, 27, 28, 29, 30, 31
                        '        y   z 
                        byte    32, 34 
  
                                       
pub main                                                  
                                                                 
  setup                                                         ' setup io and objects

  colorize_string(500)                                          ' set colors into lamps (sequential)

  term.rxflush                                                  ' clean-up

  repeat
    show_ready
    get_string                                                  ' wait for command/text input
    show_input
    check_commands                                              ' check for and run command
    if (strsize(@tbuf))                                         ' if buffer has text
      show_output
      show_string(750)                                          '  spell it out 


pub show_ready

  if (CMD_MODE == M_USB)          
    term.str(string(13, "Ready")) 


pub show_input                                                  ' for USB mode debugging

  if (CMD_MODE == M_USB)
    term.str(string(13, "<-- "))
    term.str(text.address)


pub show_output                                                 ' for USB mode debugging 

  if (CMD_MODE == M_USB)
    term.str(string(13, "--> "))
    term.str(text.address)
    
    
pub colorize_string(ms) | ch, idx, list, last, mask

'' Randomizes the colors of the string
'' -- ms is the fade-up timing
'' -- attempts to create even distribution of color table

  set_rgy_solid(OFF)

  list :=  0
  last := -1

  repeat ch from 0 to STRIP_LEN
    repeat
      idx := ||(prng.random // Colors)
      if (idx <> last)
        mask := 1 << idx
        ifnot (list & mask)
          list |= mask
          if (list == %111111)
            list := 0
          last := idx
          quit
      
    fade_up(ch, ColorTable[idx], ms)
    pixels2[ch] := pixels1[ch]                                  ' copy for re-use

  set_rgy_solid(GRN)
    

pub random_refill(ms) | ch, filled

'' Randomizes the colors of the string -- in random order!
'' -- ms is the fade-up timing
'' -- attempts to create even distribution of color table
'' -- main pixel buffer must be clear befor calling!

  set_rgy_solid(OFF)

  longfill(@pixels1, 0, STRIP_LEN)                              ' reset everything
  
  filled := 0

  repeat while (filled < STRIP_LEN)
    repeat
      ch := (prng.random >> 1) // STRIP_LEN
      if (pixels1[ch] == 0)
        fade_up(ch, pixels2[ch], ms)
        ++filled
        quit

  set_rgy_solid(GRN)
          

pub fade_up(ch, color, ms) | ustix, t, level

  ustix := ms * 1000 * US_001 / 255                             ' timing per level

  t := cnt
  repeat level from 1 to 255
    strip.setx(ch, color, level)
    waitcnt(t += ustix)


pub fade_down(ch, color, ms) | ustix, t, level

  ustix := ms * 1000 * US_001 / 255                             ' timing per level

  t := cnt
  repeat level from 254 to 0
    strip.setx(ch, color, level)
    waitcnt(t += ustix)


pub get_string | c

'' Get string from serial input stream

  text.clear

  repeat
    c := text.upper(term.rx) 
    case c
      8 :
        text.backspace
        
      13 :
        return
        
      " ", "A".."Z" :
        if (text.size < 39)
          text.putc(c)


dat { commands }

  Commands
  s_Off         byte    "OFF", 0
  s_Black       byte    "BLACK", 0
  s_Red         byte    "RED", 0
  s_Orange      byte    "ORANGE", 0
  s_Yellow      byte    "YELLOW", 0
  s_Green       byte    "GREEN", 0
  s_Cyan        byte    "CYAN", 0
  s_Blue        byte    "BLUE", 0
  s_Violet      byte    "VIOLET", 0
  s_Purple      byte    "PURPLE", 0
  s_Pink        byte    "PINK", 0
  s_White       byte    "WHITE", 0
  s_Strange     byte    "STRANGE", 0
  s_Reset       byte    "RESET", 0


pub check_commands | idx

  idx := text.str_index(@Commands, 14)                          ' look for command in list

  if (idx == -1)
    return false

  case idx
    0..1 : strip.set_all(OFF)
    2    : strip.set_all(ST_Red)
    3    : strip.set_all(ST_Orange)
    4    : strip.set_all(ST_Yellow)
    5    : strip.set_all(ST_Green)
    6    : strip.set_all(ST_Cyan)
    7    : strip.set_all(ST_Blue)
    8    : strip.set_all(ST_Violet)
    9    : strip.set_all(ST_Purple)
    10   : strip.set_all(ST_Pink)  
    11   : strip.set_all(ST_White)
    
    12 :
      strip.set_all(OFF)                                        ' black out            
      time.pause(1000)                                          ' hold                 
      random_refill(100)                                        ' re-light randomly
      term.rxflush   

    13 :
      strip.set_all(OFF)                                        ' black out            
      time.pause(1000)                                          ' hold                 
      colorize_string(100)                                      ' re-sequence colors
      term.rxflush  
      
  text.clear                                                    ' clear to prevent spelling


pub show_string(ms) | len, idx, k

  len := text.size

  repeat idx from 0 to len-1
    k := tbuf[idx]
    case k
      "A".."Z" : show_letter(k, ms)
      " "      : time.pause(ms >> 1) 


pub show_letter(c, ms) | color

  c := Alphas[c - "A"]                                          ' covert letter to channel
  color := strip.read(c)                                            
  fade_down(c, color, ms)  
  fade_up(c, color, ms)
     
                                                           
pub setup
                                                                 
'' Configure HC-8+ IO and support objects                        
                                                                 
  time.start                                                    ' setup for timing/delays
                                                                 
  prng.seed(-cnt, -cnt ~> 2, $EF8_7E4, cnt <- 2, cnt)           ' seed randomizer

  text.use(@tbuf, 40)                              
                                                                 
  io.start(0, 0)                                                ' clear pins (this cog)
                                                           
  outa[OUT7..OUT0] := %0000_0000                                ' clear outputs
  dira[OUT7..OUT0] := %1111_1111                                ' output mode

  if (PCB_REV =< "E")                                  
    io.low(OUT_EN)                                              ' enable output driver
                                                                 
  outs.start(CHANNELS, OUT0)                                    ' start outputs driver *

  strip.start_2811(@pixels1, STRIP_LEN, OUT7, 1_0)              ' start WS2811 pixel driver * 
                                                                 
  set_rs485(NOCON)                                              ' RS-485 not connected

  if (CMD_MODE == M_USB)
    term.start(RX1, TX1, %0000, USB_BAUD)                       ' use PST *
  else
    term.start(RX3, TX3, %0000, BT_BAUD)                        ' use Bluetooth *
                                                                 
  bcog := cognew(background, @bcstack) + 1                      ' start background cog *
  

con

  { --------------------------------- }
  {  S U P P O R T   R O U T I N E S  } 
  { --------------------------------- }

                                                           
pub ttl_pin                                                      
                                                                 
'' Reads TTL states for 1-of-8 input                             
'' -- returns -1 (no input)                                      
'' -- returns 0 to 7 for input on IN0 to IN7                     
''    * only one input allowed                                   
                                                                 
  if (bit_count(ttlpins) == 1)                                  ' only one button pressed?
    return bit_pos(ttlpins, 0)                                  '  yes, return button #
  else                                                           
    return -1                                                   '  none or bad input
                                                                 

pub ttl_inputs 

'' Return TTL inputs status
'' -- pins are scanned by background cog

  return ttlpins


pub dmx_address 

'' Return DMX address switch setting
'' -- switches are scanned by background cog
'' -- Note: For Rev C and lower boards, do not use DMX 256 switch -- leave it OFF

  return dmxaddr


pub read_xport | xp                                             ' Rev F and higher only
                                                                 
'' Returns state of xport bits
'' -- does not set dira bits to 0

  if (PCB_REV < "F")
    return %0000
  else
    return (ina[XP3] << 3) | (ina[XP2] << 2) | (ina[XP1] << 1) | ina[XP0]


pub write_xport(value, outsmask) | idx, mask                    ' Rev F and higher only    
                                                                 
'' Write value to xport pins                                     
'' -- outsmask used to set xport output pins                     
'' -- will not change pins with 0 bit in outsmask                
                                                                 
  if (PCB_REV < "F")                                             
    return

  ' unrolled

  if (outsmask & %1000)
    io.set(XP3, value >> 3)  

  if (outsmask & %0100)
    io.set(XP2, value >> 2)

  if (outsmask & %0010)
    io.set(XP1, value >> 1)    

  if (outsmask & %0001)
    io.set(XP0, value >> 0)

    
pub set_rgy_solid(color)                                              
                                                                  
'' Sets R/G LED                                                   
'' -- direct control only                                         
'' -- red/green/yellow + blinking requires support cog   
                                                                  
  ifnot (bcog)                                                  ' if support cog not running
    dira[R_LED..G_LED] := %11                                   ' make outputs
    case color                                                  ' set color to state                  
      OFF : outa[R_LED..G_LED] := %00                                                                 
      RED : outa[R_LED..G_LED] := %10                                                                 
      GRN : outa[R_LED..G_LED] := %01                                                                 
      YEL : outa[R_LED..G_LED] := %00                           ' yellow requires active driver       

  else                                                           
    case color                                                   
      OFF : set_rgy_flash(OFF, 100, OFF, 100)                   ' set to solid color 
      RED : set_rgy_flash(RED, 100, RED, 100)                          
      GRN : set_rgy_flash(GRN, 100, GRN, 100)                           
      YEL : set_rgy_flash(YEL, 100, YEL, 100)


pub set_rgy_flash(p0c, p0ms, p1c, p1ms)                                 
                                                                  
'' Explicit (control registers) setup of OGRY LED
'' -- requires support cog to be running               
'' -- p0c is phase 0 color (OFF, GRN, RED, YEL)                   
'' -- p0ms is phase 0 timing in milliseconds                      
'' -- p1c is phase 1 color                                        
'' -- p1ms is phase 1 timing in milliseconds

  ifnot (bcog)
    return                     
                                                                  
  if ((p0c => OFF) and (p0c =< YEL))                            ' phase 0 color
    rgycolor[0] := p0c                                           
  else                                                           
    rgycolor[0] := OFF                                           
                                                                 
  if (p0ms > 0)                                                 ' phase 0 timing                                           
    rgytime[0] := p0ms                                           
  else                                                           
    rgytime[0] := 100                                            
                                                                 
  if ((p1c => OFF) and (p1c =< YEL))                            ' phase 1 color                             
    rgycolor[1] := p1c                                           
  else                                                           
    rgycolor[1] := OFF                                           
                                                                 
  if (p1ms > 0)                                                 ' phase 1 timing                                              
    rgytime[1] := p1ms                                           
  else                                                            
    rgytime[1] := 100                                             

                                                           
pub set_rs485(state)                                              
                                                                  
'' Sets RS-485 RX and TX enable inputs                            
                                                                  
  case state                                                      
    NOCON : outa[RX2_EN..TX2_EN] := %10                           
    RX    : outa[RX2_EN..TX2_EN] := %00                           
    TX    : outa[RX2_EN..TX2_EN] := %11                           
    RXTX  : outa[RX2_EN..TX2_EN] := %01                           
                                                                  
  dira[RX2_EN..TX2_EN] := %11                                     
                                                                  
                                                                  
pub bit_count(value) : bc                                         
                                                                  
'' Returns # of bits set (1) in value

  repeat while (value)                                          ' count while there are bits
    bc += (value & %1)                                          ' add lsb bit value to count    
    value >>= 1                                                 ' next bit                                              

                                                                 
pub bit_pos(value, mode)                                         
                                                                 
'' Returns position of 1st bit set (0..31)                       
'' -- mode 0 to scan from lsb, mode 1 to scan from msb           
'' -- -1 = no bits set                                           
                                                                 
  if (value == 0)                                               ' if no bits                                   
    return -1                                                   '  return -1
  else                                                           
    if (mode == LSBFIRST)                                       ' check from LSB
      value ><= 32                                              '  flip for >|
      return (32 - >|value)                                      
    else                                                          
      return (>|value - 1)                                        
                                                                  
                                                                  
pub bit_val(value, pos)                                           
                                                            
'' Returns bit (0..1) from pos (0..31) in value                   
                                                                  
  if ((pos => 0) and (pos =< 31))                                 
    return (value >> pos) & 1                                     
  else                                                            
    return 0


con

  { ----------------------------- }
  {  B A C K G R O U N D   C O G  }
  {  - global milliseconds        }
  {  - R/G LED                    }
  {  - TTL/DMX inputs scanning    }
  { ----------------------------- }
  

var  

  long  bcog                                                    ' background cog #
  long  bcstack[64]                                             ' stack for background cog
                                                                 
  long  millis                                                  ' global milliseconds register
                                                                 
                                                                 
pri background | t                                              ' launch with cognew()
                                                                 
  io.low(R_LED)                                                 ' setup R/G LED pins
  io.low(G_LED)                                                  
                                                                 
  io.high(LD_165)                                               ' setup x165 io pins
  io.low(CLK_165)                                                
  io.input(DO_165)                                               
                                                                 
  millis := 0                                                    
                                                                 
  t := cnt                                                      ' sync loop timer
  repeat                                                         
    waitcnt(t += MS_001)                                        ' run loop every millisecond
    ++millis                                                     
    refresh_rg_led                                              ' update the R/G LED
    scan_ttl_ins                                                ' re-scan TTL & DMX inputs
                                                                 
                                                                 
var                                                              
                                                                 
  long  rgycolor[2]                                             ' led phase colors
  long  rgytime[2]                                              ' led phase timing (ms)
  long  rgphase                                                 ' current phase (red or green chip)
  long  phasetimer                                              ' time in phase
  long  rgtimer                                                 ' timer for rg process
                                                                 
                                                                 
pri refresh_rg_led                                              ' call only from support()
                                                                 
  if (++phasetimer => rgytime[rgphase])                         ' done with this phase?
    phasetimer := 0                                             ' yes, reset timer
    rgphase := 1 - rgphase                                      '  and invert phase
                                                                 
  if (++rgtimer == 16)                                           
    rgtimer := 0                                                 
                                                                 
  case rgycolor[rgphase]                                        ' set led to color for phase
    OFF:                                                         
      outa[R_LED..G_LED] := %00                                  
                                                                 
    GRN:
      outa[R_LED..G_LED] := %01  
                                                                 
    RED:
      outa[R_LED..G_LED] := %10 
                                                                   
    YEL:
      if (rgtimer < 2)                                                                       
        outa[R_LED..G_LED] := %10                               ' 1 red                        
      else                                                                                   
        outa[R_LED..G_LED] := %01                               ' 15                  


var

  long  ttlpins                                                 ' updated by background cog
  long  dmxaddr                                                  
                                                       
                                                                 
pri scan_ttl_ins | tempin                                         
                                                                 
'' Scan TTL and DMX address inputs                               
                                                                 
  outa[LD_165] := 0                                             ' blip Shift/Load line    
  outa[LD_165] := 1                                              

  ' unrolled for best speed

  tempin := (tempin << 1) | ina[DO_165]                         ' bit15
  outa[CLK_165] := 1                                            ' blip clock  
  outa[CLK_165] := 0                                                        
  tempin := (tempin << 1) | ina[DO_165]
  outa[CLK_165] := 1                   
  outa[CLK_165] := 0                   
  tempin := (tempin << 1) | ina[DO_165]
  outa[CLK_165] := 1                   
  outa[CLK_165] := 0                   
  tempin := (tempin << 1) | ina[DO_165]
  outa[CLK_165] := 1                   
  outa[CLK_165] := 0
  tempin := (tempin << 1) | ina[DO_165] 
  outa[CLK_165] := 1                    
  outa[CLK_165] := 0                    
  tempin := (tempin << 1) | ina[DO_165] 
  outa[CLK_165] := 1                    
  outa[CLK_165] := 0                    
  tempin := (tempin << 1) | ina[DO_165] 
  outa[CLK_165] := 1                    
  outa[CLK_165] := 0
  tempin := (tempin << 1) | ina[DO_165] 
  outa[CLK_165] := 1                    
  outa[CLK_165] := 0
  
  tempin := (tempin << 1) | ina[DO_165]                         ' bit7
  outa[CLK_165] := 1                    
  outa[CLK_165] := 0                    
  tempin := (tempin << 1) | ina[DO_165] 
  outa[CLK_165] := 1                    
  outa[CLK_165] := 0                    
  tempin := (tempin << 1) | ina[DO_165] 
  outa[CLK_165] := 1                    
  outa[CLK_165] := 0                    
  tempin := (tempin << 1) | ina[DO_165] 
  outa[CLK_165] := 1                    
  outa[CLK_165] := 0
  tempin := (tempin << 1) | ina[DO_165] 
  outa[CLK_165] := 1                    
  outa[CLK_165] := 0                    
  tempin := (tempin << 1) | ina[DO_165] 
  outa[CLK_165] := 1                    
  outa[CLK_165] := 0                    
  tempin := (tempin << 1) | ina[DO_165] 
  outa[CLK_165] := 1                    
  outa[CLK_165] := 0                                       
  tempin := (tempin << 1) | ina[DO_165]                         ' bit0                    
                                                         
  ttlpins := tempin.byte[0]                                     ' update global vars
  dmxaddr := tempin.byte[1] | (ina[DMX_A8] << 8)     

          
dat { license }                                                  

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

}}      